/* 
	2007-10-13 code cleanup, multi robot version ( unit id added )
	2007-10-12 ver 0.1 first run

*/

#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include "hexor.h"
#define BAUD        B115200
#define DATABITS    CS8
#define STOPBITS    1
#define PARITYON    0
#define PARITY      0
//#define 

#define FALSE 0
#define TRUE 1
#define DEVICE "/dev/ttyUSB0"
//#define DEVICE "/dev/ttyACM0"

volatile int STOP = FALSE;
volatile int ERROR = FALSE;

void signal_handler_IO(int status);	/* definition of signal handler */
int wait_flag = TRUE;		/* TRUE while no signal rece */


/**
 * Alarm signal handle. Guards serial transmission timeout. 
 */
void alarm_signal(int s)
{
    STOP = TRUE;
    ERROR = TRUE;
//    fprintf(stderr,"timeout\n");
}

/**
 * Calculates CRC rof data frame to be send to a robot.
 * New byte with CRC ia added to the end of the frame, ensure there is enough space !!
 * \param frame char * pointer to the frame
 *  Out frame description
 * frame[0] - frame length, CRC included( calculateCRC increase this by 1 )
 * frame[1] - always 1
 * frame[2] - unit number
 * frame[3] - 0
 * frame[4-] - data
 * frame[last] - crc
*/
void calculateCRC(unsigned char *frame)
{
    unsigned char crc = 0;
    int x = 0;
    int len = frame[0];
    const char crctab[256] =
	{ 0x00, 0x5E, 0xBC, 0xE2, 0x61, 0x3F, 0xDD, 0x83, 0xC2, 0x9C, 0x7E,
0x20,
	0xA3, 0xFD, 0x1F, 0x41, 0x9D, 0xC3, 0x21, 0x7F, 0xFC, 0xA2, 0x40,
	    0x1E, 0x5F, 0x01, 0xE3, 0xBD,
	0x3E, 0x60, 0x82, 0xDC, 0x23, 0x7D, 0x9F, 0xC1, 0x42, 0x1C, 0xFE,
	    0xA0, 0xE1, 0xBF, 0x5D, 0x03,
	0x80, 0xDE, 0x3C, 0x62, 0xBE, 0xE0, 0x02, 0x5C, 0xDF, 0x81, 0x63,
	    0x3D, 0x7C, 0x22, 0xC0, 0x9E,
	0x1D, 0x43, 0xA1, 0xFF, 0x46, 0x18, 0xFA, 0xA4, 0x27, 0x79, 0x9B,
	    0xC5, 0x84, 0xDA, 0x38, 0x66,
	0xE5, 0xBB, 0x59, 0x07, 0xDB, 0x85, 0x67, 0x39, 0xBA, 0xE4, 0x06,
	    0x58, 0x19, 0x47, 0xA5, 0xFB,
	0x78, 0x26, 0xC4, 0x9A, 0x65, 0x3B, 0xD9, 0x87, 0x04, 0x5A, 0xB8,
	    0xE6, 0xA7, 0xF9, 0x1B, 0x45,
	0xC6, 0x98, 0x7A, 0x24, 0xF8, 0xA6, 0x44, 0x1A, 0x99, 0xC7, 0x25,
	    0x7B, 0x3A, 0x64, 0x86, 0xD8,
	0x5B, 0x05, 0xE7, 0xB9, 0x8C, 0xD2, 0x30, 0x6E, 0xED, 0xB3, 0x51,
	    0x0F, 0x4E, 0x10, 0xF2, 0xAC,
	0x2F, 0x71, 0x93, 0xCD, 0x11, 0x4F, 0xAD, 0xF3, 0x70, 0x2E, 0xCC,
	    0x92, 0xD3, 0x8D, 0x6F, 0x31,
	0xB2, 0xEC, 0x0E, 0x50, 0xAF, 0xF1, 0x13, 0x4D, 0xCE, 0x90, 0x72,
	    0x2C, 0x6D, 0x33, 0xD1, 0x8F,
	0x0C, 0x52, 0xB0, 0xEE, 0x32, 0x6C, 0x8E, 0xD0, 0x53, 0x0D, 0xEF,
	    0xB1, 0xF0, 0xAE, 0x4C, 0x12,
	0x91, 0xCF, 0x2D, 0x73, 0xCA, 0x94, 0x76, 0x28, 0xAB, 0xF5, 0x17,
	    0x49, 0x08, 0x56, 0xB4, 0xEA,
	0x69, 0x37, 0xD5, 0x8B, 0x57, 0x09, 0xEB, 0xB5, 0x36, 0x68, 0x8A,
	    0xD4, 0x95, 0xCB, 0x29, 0x77,
	0xF4, 0xAA, 0x48, 0x16, 0xE9, 0xB7, 0x55, 0x0B, 0x88, 0xD6, 0x34,
	    0x6A, 0x2B, 0x75, 0x97, 0xC9,
	0x4A, 0x14, 0xF6, 0xA8, 0x74, 0x2A, 0xC8, 0x96, 0x15, 0x4B, 0xA9,
	    0xF7, 0xB6, 0xE8, 0x0A, 0x54,
	0xD7, 0x89, 0x6B, 0x35
    };


    frame[0]++;
    for (; x < (len); x++) {
	crc = frame[x] ^ crc;
	crc = crctab[crc];

    }
    frame[len] = crc;


}

/**
 * Sends frame to the serial link
 * Port configured by "define" directive
 * \param frame data to send with CRC
 * Return data is placed ar frame pointer. Make sure is big enough !!
*/

int sendCommand(unsigned char *frame)
{

    int len = frame[0];
    int serialfd = 0;
    int count = 0, res;
    struct termios oldtio, newtio;
    struct sigaction saio;

    if ((serialfd = open(DEVICE, O_RDWR | O_NOCTTY | O_NONBLOCK)) == -1) {
	fprintf(stderr, "Unable to open serial port, check config!!!");
	exit(1);
    }
    STOP = FALSE;
    tcgetattr(serialfd, &oldtio);	// save current port settings 
    // set new port settings for canonical input processing 
    bzero(&newtio, sizeof(newtio));
  newtio.c_cflag = CS8 | CREAD | CLOCAL | IXOFF | BRKINT;
  newtio.c_iflag = IGNPAR;
  newtio.c_oflag = 0;
  newtio.c_lflag = 0;
  newtio.c_cc[VMIN] = 1;
  newtio.c_cc[VTIME] = 0;
  newtio.c_ospeed = BAUD;
  newtio.c_ispeed = BAUD;



    bzero(&saio, sizeof(saio));
    saio.sa_handler = signal_handler_IO;
    saio.sa_flags = 0;
    saio.sa_restorer = NULL;
    sigaction(SIGIO, &saio, NULL);

    fcntl(serialfd, F_SETOWN, getpid());
    fcntl(serialfd, F_SETFL, FASYNC);

    tcflush(serialfd, TCIFLUSH);
    tcsetattr(serialfd, TCSANOW, &newtio);

    count = write(serialfd, frame, len);
    if (count != len) {
	fprintf(stderr, "Write error");
	exit(1);
    }

    /* wait 2s for response */
    signal(SIGALRM, alarm_signal);
    alarm(2);

    res = 0;
    while (STOP == FALSE) {
	usleep(50000);
	if (wait_flag == FALSE) {
	    if (res == 0) {
		res = read(serialfd, frame, 1);
	    }
	    res += read(serialfd, (frame + res), 255 - res);
	    if (res == frame[0]) {
		STOP = TRUE;
		ERROR = FALSE;
	    }
	    wait_flag = TRUE;
	}
    }

    tcflush(serialfd, TCIFLUSH);
    tcsetattr(serialfd, TCSANOW, &oldtio);

    signal(SIGALRM, SIG_IGN);
    close(serialfd);

    if (ERROR)
	return 1;
    else
	return 0;
}

/**
 * IO signal handler. sets wait_flag to FALSE, to indicate above loop that 
 * characters have been received
*/

void signal_handler_IO(int status)
{

    signal(SIGALRM, SIG_IGN);
//    fprintf(stderr,"received SIGIO signal.\n");
    wait_flag = FALSE;
}


/**
 * Prints frame.
 * \param frame Frame to print
*/

void printFrame(unsigned char *frame)
{
    int i;
    for (i = 0; i < frame[0]; i++) {
	printf("0x%02x ", frame[i]);
    }
    printf("\n");
}


/** 
 * Go forward
 * \param unit robot id
*/
int forward(unsigned char unit)
{
    unsigned char frame[10] = { 0x06, 0x01, 0x0b, 0x00, 0x20, 0x01 };
    frame[2] = unit;
    calculateCRC(frame);
    return sendCommand(frame);



}

/** 
 * Go backward
 * \param unit robot id
*/
int backward(unsigned char unit)
{
    unsigned char frame[10] = { 0x06, 0x01, 0x0b, 0x00, 0x20, 0x04 };
    frame[2] = unit;
    calculateCRC(frame);
//    printFrame(frame);
    return sendCommand(frame);

}

/** 
 * Turn left
 * \param unit robot id
*/

int tleft(unsigned char unit)
{
    unsigned char frame[10] = { 0x06, 0x01, 0x0b, 0x00, 0x20, 0x02 };
    frame[2] = unit;
    calculateCRC(frame);
//    printFrame(frame);
    return sendCommand(frame);


}

/** 
 * Turn right
 * \param unit robot id
*/

int tright(unsigned char unit)
{
    unsigned char frame[10] = { 0x06, 0x01, 0x0b, 0x00, 0x20, 0x03 };
    frame[2] = unit;
    calculateCRC(frame);
//    printFrame(frame);
    return sendCommand(frame);


}

/** 
 * Stop
 * \param unit robot id
*/

int stop(unsigned char unit)
{
    unsigned char frame[10] = { 0x06, 0x01, 0x0b, 0x00, 0x20, 0x00 };
    frame[2] = unit;
    calculateCRC(frame);
//    printFrame(frame);
    return sendCommand(frame);
}

/** 
 * Set camera position
 * \param unit robot id
 * \param x horizontal position
 * \param y vertical position
 * x,y params range 0x44-0xf5
*/
int camera(unsigned char unit, long x, long y)
{
    int res;
    unsigned char framex[10] =
	{ 0x07, 0x01, 0x0b, 0x00, 0x34, 0x04, 0x00 };
    unsigned char framey[10] =
	{ 0x07, 0x01, 0x0b, 0x00, 0x34, 0x05, 0x00 };
    framex[2] = unit;
    framey[2] = unit;
    framex[6] = x;
    framey[6] = y;
    calculateCRC(framex);
    calculateCRC(framey);
//        printFrame(frame);
    res = sendCommand(framex);
    res = sendCommand(framey);

//        printFrame(frame);
    return res;
}


/** 
 * Turn camera on and off
 * \param unit robot id
 * \parma state new camera state ( true -> on , false -> off )
 *
*/
int cameraOnOff(unsigned char unit, int state)
{
    int res;
    unsigned char frame[10] =
	{ 0x06, 0x01, 0x0b, 0x00, 0x24, 0x00 };
    frame[2] = unit;
    if ( state ) frame[5] = 0x02; else frame[5]=0x00;
    calculateCRC(frame);
//        printFrame(frame);
    res = sendCommand(frame);

//        printFrame(frame);
    return res;
}



/** 
 * Read sonar and temperature readings
 * \param unit robot id
 * \param temp temperature
 * \param sonar distance in some units ;)
*/
int read_sensors(unsigned char unit, long *temp, long *sonar)
{
    int ret;
    unsigned char frame[12] = { 0x05, 0x01, 0x0b, 0x00, 0x21 };
    frame[2] = unit;
    calculateCRC(frame);
    ret = sendCommand(frame);
    *temp = frame[5] << 8 | frame[6];
    *sonar = frame[7] << 8 | frame[8];
    return ret;
}


/** 
 * Read IR and mechanical sensors
 * \param unit robot id
 * \param ir sensors state ( activated sensor = zeroed bit )
 * 
*/

int read_ir(unsigned char unit, unsigned char *ir)
{
    int ret;
    unsigned char frame[12] = { 0x05, 0x01, 0x0b, 0x00, 0x27 };
    frame[2] = unit;
    calculateCRC(frame);
    // printFrame(frame);
    ret = sendCommand(frame);
    // printFrame(frame);
    *ir = frame[6];
    return ret;

}




/* TODO
	1. camera on/off
	2. cleanup serial setup code
*/

/*
int main(){
char x;
//       camera(0xf5,0xf4);
 //      sleep(1);
for (x=0;x<15;x++){
	

	read_ir(11,&x);
        printf("ir: 0%X\n",x);
	usleep(500000);
}

}
*/
