#include <stdio.h>
#include <epp.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>
#include "hexor.h"


/** 
 * delay predicate
 * \param a <b>in</b> dealy in seconds
*/
pl_predicate delay_predicate(pl_term a)
{
    long d;

    if (pl_term2long(&d, a) == PL_TRUE) {
	if (usleep(d*1000000) == 0);
	return (PL_TRUE);
    }

    return (PL_FALSE);
}

/** 
 * read sensors predicate
 * \param u <b>in</b> unit number
 * \param a <b>out</b> temperature readings
 * \param b <b>out</b> sonar readings
*/

pl_predicate read_sensors_predicate(pl_term u, pl_term a, pl_term b)
{
    long temp, sonar;
    long unit;
    if (pl_term2long(&unit, u) == PL_TRUE) {
	if (read_sensors(unit, &temp, &sonar) == 0 && pl_longUterm(a, temp)
	    && pl_longUterm(b, sonar)) {
	    return (PL_TRUE);
	}
    }
    return (PL_FALSE);
}


/** 
 * read ir & touch sensors predicate
 * \param u <b>in</b> unit number
 * \param a <b>out</b> sensor state ( binary )
 * TODO: bit names
*/
pl_predicate read_ir_predicate(pl_term u, pl_term a)
{
    long iv;
    long unit;
    if (pl_term2long(&unit, u) == PL_TRUE) {
	if (read_ir((unsigned char)unit, (unsigned char *)&iv) == 0 && pl_longUterm(a, iv)) {
	    return (PL_TRUE);
	}
    }
    return (PL_FALSE);
}


/** 
 * set camera & sonar position
 * \param u <b>in</b> unit number
 * \param a <b>out</b> horizontal position
 * \param b <b>out</b> vertical position
 * Valid range 0x44-0xf5.
*/
pl_predicate camera_predicate(pl_term u, pl_term a, pl_term b)
{
    long x, y;
    long unit;
    if (pl_term2long(&unit, u) == PL_TRUE) {
	if (pl_term2long(&x, a) && pl_term2long(&y, b)) {
	    if (camera(unit, x, y) == 0)
		return (PL_TRUE);
	    return (PL_TRUE);
	}
    }
    return (PL_FALSE);

}

/** 
 * Camera on predicate
 * \param u <b>in</b> unit number
*/
pl_predicate camera_on_predicate(pl_term u)
{
    long unit;
    if (pl_term2long(&unit, u) == PL_TRUE) {
	if (cameraOnOff(unit,1) == 0)
	    return (PL_TRUE);
    }
    return (PL_FALSE);

}


/** 
 * Camera off predicate
 * \param u <b>in</b> unit number
*/
pl_predicate camera_off_predicate(pl_term u)
{
    long unit;
    if (pl_term2long(&unit, u) == PL_TRUE) {
	if (cameraOnOff(unit,0) == 0)
	    return (PL_TRUE);
    }
    return (PL_FALSE);

}


/** 
 * Go forward predicate
 * \param u <b>in</b> unit number
*/
pl_predicate forward_predicate(pl_term u)
{
    long unit;
    if (pl_term2long(&unit, u) == PL_TRUE) {
	if (forward(unit) == 0)
	    return (PL_TRUE);
    }
    return (PL_FALSE);

}

/** 
 * Go backward predicate
 * \param u <b>in</b> unit number
*/
pl_predicate backward_predicate(pl_term u)
{
    long unit;
    if (pl_term2long(&unit, u) == PL_TRUE) {
	if (backward(unit) == 0)
	    return (PL_TRUE);
    }
    return (PL_FALSE);

}

/** 
 * Turn left predicate
 * \param u <b>in</b> unit number
*/
pl_predicate left_predicate(pl_term u)
{

    long unit;
    if (pl_term2long(&unit, u) == PL_TRUE) {
	if (tleft(unit) == 0)
	    return (PL_TRUE);
    }
    return (PL_FALSE);
}

/** 
 * Turn right predicate
 * \param u <b>in</b> unit number
*/
pl_predicate right_predicate(pl_term u)
{

    long unit;
    if (pl_term2long(&unit, u) == PL_TRUE) {
	if (tright(unit) == 0)
	    return (PL_TRUE);
    }
    return (PL_FALSE);
}

/** 
 * Stop predicate
 * \param u <b>in</b> unit number
*/
pl_predicate stop_predicate(pl_term u)
{

    long unit;
    if (pl_term2long(&unit, u) == PL_TRUE) {
	if (stop(unit) == 0)
	    return (PL_TRUE);
    }
    return (PL_FALSE);
}




void *thread(void *unused) {
    pl_attach_engine();

//    pl_console();
    pl_call("runme2", NULL);


    pl_detach_engine();
    return(NULL);
}

int main(int argc, char *argv[])
{
    pthread_t t;
    pl_register_predicate("forward", 1, forward_predicate);
    pl_register_predicate("backward", 1, backward_predicate);
    pl_register_predicate("left", 1, left_predicate);
    pl_register_predicate("right", 1, right_predicate);
    pl_register_predicate("stop", 1, stop_predicate);
    pl_register_predicate("delay", 1, delay_predicate);
    pl_register_predicate("readSensors", 3, read_sensors_predicate);
    pl_register_predicate("readIR", 2, read_ir_predicate);
    pl_register_predicate("camera", 3, camera_predicate);
    pl_register_predicate("cameraOn", 1, camera_on_predicate);
    pl_register_predicate("cameraOff", 1, camera_off_predicate);
    pl_init_from_state( *argv);
    
    pthread_create(&t, NULL, thread, NULL);
    
    // pl_console();
    pl_call("runme1", NULL);

    pthread_join(t, NULL);

    pl_shutdown();
    return (0);
}
